""" Simple replacement for distlib SimpleScrapingLocator """
import logging

import distlib.locators
import requests
from distlib.locators import SimpleScrapingLocator

from .util import TimedCache

LOG = logging.getLogger(__name__)


class SimpleJsonLocator(object):

    """Simple replacement for distlib SimpleScrapingLocator"""

    def __init__(self, base_index):
        self.base_index = base_index
        # 16m cache
        self._cache = TimedCache(1000, self._get_releases)

    def get_releases(self, project_name):
        return self._cache[project_name]

    def _get_releases(self, project_name):
        url = "%s/pypi/%s/json" % (self.base_index, project_name)
        response = requests.get(url)
        try:
            response.raise_for_status()
        except requests.HTTPError as e:
            LOG.warning("Error fetching '%s' from upstream: %s", project_name, e)
            return []
        data = response.json()
        items = []
        summary = data["info"].get("summary")
        for version, releases in data["releases"].items():
            for release in releases:
                try:
                    item = {
                        "name": project_name,
                        "version": version,
                        "summary": summary,
                        "url": release["url"],
                        "digests": release.get("digests", {}),
                        "requires_python": release["requires_python"],
                    }
                except KeyError:
                    continue
                items.append(item)
        return items


class FormattedScrapingLocator(SimpleScrapingLocator):
    def get_releases(self, project_name):
        projects = self.get_project(project_name)
        items = []
        for version, urls in projects["urls"].items():
            for url in urls:
                dist = projects[version]
                digest = projects["digests"].get(url)
                digest_dict = {}
                if digest is not None:
                    digest_dict[digest[0]] = digest[1]
                items.append(
                    {
                        "name": dist.name,
                        "version": dist.version,
                        "summary": dist.metadata.dictionary.get("summary"),
                        "url": url,
                        "digests": digest_dict,
                        "requires_python": dist.metadata.dictionary.get(
                            "requires_python"
                        ),
                    }
                )
        return items


# Distlib checks if wheels are compatible before returning them.
# This is useful if you are attempting to install on the system running
# distlib, but we actually want ALL wheels so we can display them to the
# clients.  So we have to monkey patch the method. I'm sorry.
def is_compatible(wheel, tags=None):
    """Hacked function to monkey patch into distlib"""
    return True


distlib.locators.is_compatible = is_compatible



""" Helpers for syncing packages into the cache in AWS Lambda """
import argparse
import json
import logging
import os
import shutil
import subprocess
import sys
import tempfile
import zipfile
from distutils.spawn import find_executable  # pylint: disable=E0611,F0401
from urllib.request import urlretrieve

import boto3
from pyramid.paster import get_appsettings

from pypicloud import __version__, _lambda_handler
from pypicloud.storage.s3 import S3Storage

HANDLER_FILENAME = "lambda_script.py"

VENV_VERSION = "15.1.0"
VENV_URL = (
    "https://pypi.org/packages/source/v/"
    "virtualenv/virtualenv-%s.tar.gz" % VENV_VERSION
)


def _create_role(role_name, description, policy):
    """Idempotently create the IAM Role"""
    iam = boto3.client("iam")
    iam_resource = boto3.resource("iam")
    role = iam_resource.Role(role_name)
    try:
        role.load()
    except iam.exceptions.NoSuchEntityException:
        print("Creating IAM role %s" % role_name)
        iam.create_role(
            Path="/pypicloud/",
            RoleName=role_name,
            AssumeRolePolicyDocument="""{
            "Version": "2012-10-17",
            "Statement": [{
            "Effect": "Allow",
            "Principal": {"Service": ["lambda.amazonaws.com"]},
            "Action": ["sts:AssumeRole"]
            }]}
            """,
            Description=description,
        )
        role.load()

    policy_name = "inline_policy"
    exists = False
    for role_policy in role.policies.all():
        if role_policy.name == policy_name:
            exists = True
            break
    if not exists:
        print("Attaching inline role policy")
        iam.put_role_policy(
            RoleName=role_name, PolicyName=policy_name, PolicyDocument=policy
        )
    return role.arn


def _create_dynamodb_role(settings, bucket):
    """Create the AWS Role needed for Lambda with a DynamoDB cache"""
    # Jump through hoops to get the table names
    from pypicloud.cache.dynamo import DynamoPackage, PackageSummary

    namespace = settings.get("db.namespace", ())
    policy = """{
        "Version": "2012-10-17",
        "Statement": [
            {
                "Effect": "Allow",
                "Action": "logs:*",
                "Resource": "arn:aws:logs:*"
            },
            {
                "Effect": "Allow",
                "Action": [
                    "s3:GetObject"
                ],
                "Resource": "arn:aws:s3:::%s/*"
            },
            {
                "Effect": "Allow",
                "Action": [
                    "dynamodb:ListTables"
                ],
                "Resource": "*"
            },
            {
                "Effect": "Allow",
                "Action": [
                    "dynamodb:BatchGetItem",
                    "dynamodb:BatchWriteItem",
                    "dynamodb:DeleteItem",
                    "dynamodb:DescribeLimits",
                    "dynamodb:DescribeTable",
                    "dynamodb:GetItem",
                    "dynamodb:GetRecords",
                    "dynamodb:PutItem",
                    "dynamodb:Query",
                    "dynamodb:Scan",
                    "dynamodb:UpdateItem"
                ],
                "Resource": [%s]
            }]}
    """
    dynamo = boto3.client("dynamodb", region_name=settings["db.region_name"])
    table_arns = []
    for model in (DynamoPackage, PackageSummary):
        tablename = model.meta_.ddb_tablename(namespace)
        try:
            result = dynamo.describe_table(TableName=tablename)
        except dynamo.exceptions.ResourceNotFoundException:
            print(
                "Could not find DynamoDB table %r. "
                "Try running the server first to create the tables" % tablename
            )
            sys.exit(1)
        table_arns.append(result["Table"]["TableArn"])
    index_arns = [t + "/index/*" for t in table_arns]
    table_arns.extend(index_arns)
    table_arns = ['"' + t + '"' for t in table_arns]
    full_policy = policy % (bucket.name, (",".join(table_arns)))

    return _create_role(
        "pypicloud_lambda_dynamo",
        "Lambda role for syncing S3 package changes to DynamoDB",
        full_policy,
    )


def _create_default_role(settings, bucket):
    """Create the AWS Role needed for Lambda with most caches"""
    policy = (
        """{
        "Version": "2012-10-17",
        "Statement": [
            {
                "Effect": "Allow",
                "Action": "logs:*",
                "Resource": "arn:aws:logs:*"
            },
            {
                "Effect": "Allow",
                "Action": [
                    "s3:GetObject"
                ],
                "Resource": "arn:aws:s3:::%s/*"
            }]}
    """
        % bucket.name
    )
    db = settings["pypi.db"]
    return _create_role(
        "pypicloud_lambda_" + db,
        "Lambda role for syncing S3 package changes to " + db,
        policy,
    )


def make_virtualenv(env):
    """Create a virtualenv"""
    if find_executable("virtualenv") is not None:
        cmd = ["virtualenv"] + [env]
        subprocess.check_call(cmd)
    else:
        # Otherwise, download virtualenv from pypi
        path = urlretrieve(VENV_URL)[0]
        try:
            subprocess.check_call(["tar", "xzf", path])
            subprocess.check_call(
                [sys.executable, "virtualenv-%s/virtualenv.py" % VENV_VERSION, env]
            )
        finally:
            os.unlink(path)
            shutil.rmtree("virtualenv-%s" % VENV_VERSION)


def build_lambda_bundle(argv=None):
    """Build the zip bundle that will be deployed to AWS Lambda"""
    if argv is None:
        argv = sys.argv[1:]
    parser = argparse.ArgumentParser(description=build_lambda_bundle.__doc__)
    parser.add_argument(
        "-o", help="Name of output file (default %(default)s)", default="output.zip"
    )
    parser.add_argument("config", help="Name of config file")

    args = parser.parse_args(argv)
    settings = get_appsettings(args.config)
    bundle = _build_lambda_bundle(settings)
    with open(args.o, "wb") as ofile:
        ofile.write(bundle)
    print("Wrote bundle file %s" % args.o)


def _build_lambda_bundle(settings):
    """Build the lambda bundle"""
    venv_dir = tempfile.mkdtemp()
    try:
        print("Creating virtualenv %s" % venv_dir)
        make_virtualenv(venv_dir)
        print("Installing pypicloud into virtualenv")
        pip = os.path.join(venv_dir, "bin", "pip")
        subprocess.check_call(
            [pip, "install", "pypicloud[dynamo,redis]==%s" % __version__]
        )

        bundle = "%s/output.zip" % venv_dir
        zipf = zipfile.ZipFile(bundle, "w", zipfile.ZIP_DEFLATED)

        site_packages = os.path.join(venv_dir, "lib", "site-packages")
        if not os.path.exists(site_packages):
            subdir = [
                f
                for f in os.listdir(os.path.join(venv_dir, "lib"))
                if f.startswith("python")
            ][0]
            site_packages = os.path.join(venv_dir, "lib", subdir, "site-packages")

        # Hack because importing some modules blows up on Lambda
        for mod in ["zope", "repoze"]:
            open(os.path.join(site_packages, mod, "__init__.py"), "a").close()
        for root, dirs, files in os.walk(site_packages):
            for filename in files:
                fullpath = os.path.join(root, filename)
                zipf.write(fullpath, os.path.relpath(fullpath, site_packages))

        handler = os.path.join(
            os.path.dirname(_lambda_handler.__file__), "_lambda_handler.py"
        )
        zipf.write(handler, HANDLER_FILENAME)

        zipf.close()
        with open(bundle, "rb") as ifile:
            return ifile.read()
    finally:
        shutil.rmtree(venv_dir)


def create_sync_scripts(argv=None):
    """
    Set bucket notifications and create AWS Lambda functions that will sync
    changes in the S3 bucket to the cache

    """
    if argv is None:
        argv = sys.argv[1:]
    parser = argparse.ArgumentParser(description=create_sync_scripts.__doc__)
    parser.add_argument(
        "-n",
        help="Name of the AWS Lambda function to " " create (default %(default)s)",
        default="pypicloud_s3_sync",
    )
    parser.add_argument(
        "-f", help="Overwrite existing AWS Lambda functions", action="store_true"
    )
    parser.add_argument(
        "-a",
        help="The ARN of the IAM role to run as. If not "
        "provided, pypicloud will attempt to create one.",
    )
    parser.add_argument(
        "-r",
        help="The AWS region to create the Lambda in "
        "(will attempt to auto-detect from config file)",
    )
    parser.add_argument("config", help="Name of config file")

    args = parser.parse_args(argv)
    logging.basicConfig()
    settings = get_appsettings(args.config)

    if args.r:
        region = args.r
    else:
        if settings.get("pypi.storage") == "s3":
            region = settings.get("storage.region_name")
    if not region:
        print("No region detected in config file. Please use -r <region>")
        sys.exit(1)

    kwargs = S3Storage.configure(settings)
    bucket = kwargs["bucket"]

    # Create the Role
    if args.a:
        role_arn = args.a
    else:
        db = settings["pypi.db"]
        if db == "dynamo":
            role_arn = _create_dynamodb_role(settings, bucket)
        else:
            role_arn = _create_default_role(settings, bucket)

    lam = boto3.client("lambda", region_name=region)
    func_arn = None
    try:
        func = lam.get_function(FunctionName=args.n)
    except lam.exceptions.ResourceNotFoundException:
        pass
    else:
        if args.f:
            print("Deleting pre-existing function %r" % args.n)
            lam.delete_function(FunctionName=args.n)
        else:
            func_arn = func["Configuration"]["FunctionArn"]
            print("Lambda function %r already exists. Use -f to overwrite" % args.n)

    # Create the lambda function
    if func_arn is None:
        bundle = _build_lambda_bundle(settings)
        handler_module = os.path.splitext(HANDLER_FILENAME)[0]
        # Pull out only the cache db settings
        small_settings = {"pypi.db": settings["pypi.db"]}
        for key, val in settings.items():
            if key.startswith("db."):
                small_settings[key] = val

        print("Creating Lambda function %s" % args.n)
        func = lam.create_function(
            FunctionName=args.n,
            Runtime="python2.7",
            Handler=handler_module + ".handle_s3_event",
            Code={"ZipFile": bundle},
            Environment={
                "Variables": {"PYPICLOUD_SETTINGS": json.dumps(small_settings)}
            },
            Description="Process S3 Object notifications & update pypicloud cache",
            Timeout=30,
            Publish=True,
            Role=role_arn,
        )
        func_arn = func["FunctionArn"]

    print("Adding invoke permission for S3")
    account_id = boto3.client("sts").get_caller_identity().get("Account")
    try:
        lam.add_permission(
            Action="lambda:InvokeFunction",
            FunctionName=args.n,
            Principal="s3.amazonaws.com",
            SourceAccount=account_id,
            SourceArn="arn:aws:s3:::" + bucket.name,
            StatementId="s3_invoke",
        )
    except lam.exceptions.ResourceConflictException:
        print("Permission already present")

    print("Adding lambda configuration to S3 bucket")
    notification = bucket.Notification()
    notification.put(
        NotificationConfiguration={
            "LambdaFunctionConfigurations": [
                {"LambdaFunctionArn": func_arn, "Events": ["s3:ObjectCreated:*"]},
                {"LambdaFunctionArn": func_arn, "Events": ["s3:ObjectRemoved:*"]},
            ]
        }
    )



""" Store package data in a SQL database """
import json
import logging

import zope.sqlalchemy
from pyramid.settings import asbool
from sqlalchemy import (
    Column,
    DateTime,
    String,
    Text,
    and_,
    distinct,
    engine_from_config,
    or_,
)
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.ext.mutable import Mutable
from sqlalchemy.orm import sessionmaker
from sqlalchemy.sql import func
from sqlalchemy.types import TEXT, TypeDecorator

from pypicloud.dateutil import UTC, utcnow
from pypicloud.models import Package

from .base import ICache

LOG = logging.getLogger(__name__)

Base = declarative_base()  # pylint: disable=C0103


class JSONEncodedDict(TypeDecorator):  # pylint: disable=W0223

    "Represents an immutable structure as a json-encoded string."

    impl = TEXT

    def process_bind_param(self, value, dialect):
        if value is not None:
            value = json.dumps(value)
        return value

    def process_result_value(self, value, dialect):
        if value is not None:
            value = json.loads(value)
        return value


class MutableDict(Mutable, dict):

    """SQLAlchemy dict field that tracks changes"""

    @classmethod
    def coerce(cls, key, value):
        "Convert plain dictionaries to MutableDict."

        if not isinstance(value, MutableDict):
            if isinstance(value, dict):
                return MutableDict(value)

            # this call will raise ValueError
            return Mutable.coerce(key, value)
        else:
            return value

    def __setitem__(self, key, value):
        "Detect dictionary set events and emit change events."

        dict.__setitem__(self, key, value)
        self.changed()

    def __delitem__(self, key):
        "Detect dictionary del events and emit change events."

        dict.__delitem__(self, key)
        self.changed()


MutableDict.associate_with(JSONEncodedDict)


class TZAwareDateTime(TypeDecorator):  # pylint: disable=W0223
    impl = DateTime
    cache_ok = True

    def process_bind_param(self, value, dialect):
        if value.tzinfo is None:
            return value.replace(tzinfo=UTC)
        return value

    def process_result_value(self, value, dialect):
        if value.tzinfo is None:
            return value.replace(tzinfo=UTC)
        return value


class SQLPackage(Package, Base):

    """Python package stored in SQLAlchemy"""

    __tablename__ = "packages"
    filename = Column(String(255, convert_unicode=True), primary_key=True)
    name = Column(String(255, convert_unicode=True), index=True, nullable=False)
    version = Column(String(1000, convert_unicode=True), nullable=False)
    last_modified = Column(TZAwareDateTime(), index=True, nullable=False)
    # TEXT, as pypi does the same, and using String(N) would mismatch with pypi
    summary = Column(Text(), nullable=True)
    data = Column(JSONEncodedDict(), nullable=False)


def create_schema(engine):
    """
    Create the database schema if needed

    Parameters
    ----------
    engine : :class:`sqlalchemy.Engine`

    Notes
    -----
    The method should only be called after importing all modules containing
    models which extend the ``Base`` object.

    """
    Base.metadata.create_all(bind=engine)


def drop_schema(engine):
    """
    Drop the database schema

    Parameters
    ----------
    engine : :class:`sqlalchemy.Engine`

    Notes
    -----
    The method should only be called after importing all modules containing
    models which extend the ``Base`` object.

    """
    Base.metadata.drop_all(bind=engine)


class SQLCache(ICache):

    """Caching database that uses SQLAlchemy"""

    def __init__(self, request=None, dbmaker=None, graceful_reload=False, **kwargs):
        super(SQLCache, self).__init__(request, **kwargs)
        self.dbmaker = dbmaker
        self.db = self.dbmaker()
        self.graceful_reload = graceful_reload

        if request is not None:
            zope.sqlalchemy.register(self.db, transaction_manager=request.tm)

    def new_package(self, *args, **kwargs):
        return SQLPackage(*args, **kwargs)

    def reload_if_needed(self):
        super(SQLCache, self).reload_if_needed()
        if self.request is None:
            self.db.commit()
            self.db.close()

    @classmethod
    def configure(cls, settings):
        kwargs = super(SQLCache, cls).configure(settings)
        graceful_reload = asbool(settings.pop("db.graceful_reload", False))
        settings.read_prefix_from_environ("db.")
        engine = engine_from_config(settings, prefix="db.")
        # Create SQL schema if not exists
        create_schema(engine)
        kwargs["dbmaker"] = sessionmaker(bind=engine)
        kwargs["graceful_reload"] = graceful_reload
        return kwargs

    @classmethod
    def postfork(cls, **kwargs):
        # Have to dispose of connections after uWSGI forks,
        # otherwise they'll get corrupted.
        kwargs["dbmaker"].kw["bind"].dispose()

    def fetch(self, filename):
        return self.db.query(SQLPackage).filter_by(filename=filename).first()

    def all(self, name):
        pkgs = self.db.query(SQLPackage).filter_by(name=name).all()
        pkgs.sort(reverse=True)
        return pkgs

    def distinct(self):
        names = self.db.query(distinct(SQLPackage.name)).order_by(SQLPackage.name).all()
        return [n[0] for n in names]

    def search(self, criteria, query_type):
        """
        Perform a search.

        Queries are performed as follows:

            For the AND query_type, queries within a column will utilize the
            AND operator, but will not conflict with queries in another column.

                (column1 LIKE '%a%' AND column1 LIKE '%b%')
                OR
                (column2 LIKE '%c%' AND column2 LIKE '%d%')

            For the OR query_type, all queries will utilize the OR operator:

                (column1 LIKE '%a%' OR column1 LIKE '%b%')
                OR
                (column2 LIKE '%c%' OR column2 LIKE '%d%')

        """
        conditions = []
        for key, queries in criteria.items():
            # Make sure search key exists in the package class.
            # It should be either "name" or "summary".
            field = getattr(SQLPackage, key, None)
            if not field:
                continue

            column_conditions = []

            for query in queries:
                # Generate condition and add to list
                condition = field.like("%" + query + "%")
                column_conditions.append(condition)

            # Check if conditions for this column were generated
            if column_conditions:
                if query_type == "and":
                    operator = and_
                else:
                    operator = or_
                conditions.append(operator(*column_conditions))

        # Piece together the queries. Refer to the method docstring for
        # examples as to how this works.
        results = self.db.query(SQLPackage).filter(or_(*conditions))

        # Extract only the most recent version for each package
        latest_map = {}
        for package in results.all():
            if package.name not in latest_map or package > latest_map[package.name]:
                latest_map[package.name] = package

        return latest_map.values()

    def summary(self):
        subquery = (
            self.db.query(
                SQLPackage.name,
                func.max(SQLPackage.last_modified).label("last_modified"),
            )
            .group_by(SQLPackage.name)
            .subquery()
        )
        rows = self.db.query(
            SQLPackage.name, SQLPackage.last_modified, SQLPackage.summary
        ).filter(
            (SQLPackage.name == subquery.c.name)
            & (SQLPackage.last_modified == subquery.c.last_modified)
        )

        # Dedupe because two packages may share the same last_modified
        seen_packages = set()
        packages = []
        for row in rows:
            if row[0] in seen_packages:
                continue
            seen_packages.add(row[0])
            packages.append(
                {"name": row[0], "last_modified": row[1], "summary": row[2]}
            )
        return packages

    def clear(self, package):
        self.db.delete(package)

    def clear_all(self):
        # Release any transactions before we go reloading schema
        if self.request is None:
            self.db.rollback()
        else:
            self.request.tm.abort()
        engine = self.dbmaker.kw["bind"]
        drop_schema(engine)
        create_schema(engine)

    def save(self, package):
        self.db.merge(package)

    def reload_from_storage(self, clear=True):
        if not self.graceful_reload:
            return super(SQLCache, self).reload_from_storage(clear)

        LOG.info("Rebuilding cache from storage")
        # Log start time
        start = utcnow()
        # Fetch packages from storage s1
        s1 = set(self.storage.list(SQLPackage))
        # Fetch cache packages c1
        c1 = set(self.db.query(SQLPackage).all())
        # Add missing packages to cache (s1 - c1)
        missing = s1 - c1
        if missing:
            LOG.info("Adding %d missing packages to cache", len(missing))
            for pkg in missing:
                self.db.merge(pkg)
        # Delete extra packages from cache (c1 - s1) when last_modified < start
        # The time filter helps us avoid deleting packages that were
        # concurrently uploaded.
        extra1 = [p for p in (c1 - s1) if p.last_modified <= start]
        if extra1:
            LOG.info("Removing %d extra packages from cache", len(extra1))
            for pkg in extra1:
                self.db.query(SQLPackage).filter(
                    SQLPackage.filename == pkg.filename
                ).delete(synchronize_session=False)

        # If any packages were concurrently deleted during the cache rebuild,
        # we can detect them by polling storage again and looking for any
        # packages that were present in s1 and are missing from s2
        s2 = set(self.storage.list(SQLPackage))
        # Delete extra packages from cache (s1 - s2)
        extra2 = s1 - s2
        if extra2:
            LOG.info(
                "Removing %d packages from cache that were concurrently "
                "deleted during rebuild",
                len(extra2),
            )
            for pkg in extra2:
                self.db.query(SQLPackage).filter(
                    SQLPackage.filename == pkg.filename
                ).delete(synchronize_session=False)

    def check_health(self):
        try:
            self.db.query(SQLPackage).first()
        except SQLAlchemyError as e:
            return (False, str(e))
        else:
            return (True, "")



""" Commandline scripts """
import argparse
import getpass
import gzip
import json
import logging
import os
import sys
from base64 import b64encode

import transaction
from jinja2 import Template
from pkg_resources import resource_string  # pylint: disable=E0611
from pyramid.paster import bootstrap

from pypicloud.access import SCHEMES, get_pwd_context


def gen_password(argv=None):
    """Generate a salted password"""
    if argv is None:
        argv = sys.argv[1:]
    parser = argparse.ArgumentParser(gen_password.__doc__)
    parser.add_argument("-i", help="Read password from stdin", action="store_true")
    parser.add_argument("-r", help="Number of rounds", type=int)
    parser.add_argument(
        "-s",
        help="Hashing scheme (default %(default)s)",
        default=SCHEMES[0],
        choices=SCHEMES,
    )
    args = parser.parse_args(argv)
    if args.i:
        password = sys.stdin.readline()
    else:
        password = _get_password()
    print(_gen_password(password, args.s, args.r))


def _get_password():
    """Prompt user for a password twice for safety"""
    while True:
        password = getpass.getpass()
        verify = getpass.getpass()
        if password == verify:
            return password
        else:
            print("Passwords do not match!")


def _gen_password(password: str, scheme: str = None, rounds: int = None) -> str:
    pwd_context = get_pwd_context(scheme, rounds)
    return pwd_context.hash(password)


NO_DEFAULT = object()


def wrapped_input(msg):
    """Wraps input for tests"""
    return input(msg)


def prompt(msg, default=NO_DEFAULT, validate=None):
    """Prompt user for input"""
    while True:
        response = wrapped_input(msg + " ").strip()
        if not response:
            if default is NO_DEFAULT:
                continue
            return default
        if validate is None or validate(response):
            return response


def prompt_option(text, choices, default=NO_DEFAULT):
    """Prompt the user to choose one of a list of options"""
    while True:
        for i, msg in enumerate(choices):
            print("[%d] %s" % (i + 1, msg))
        response = prompt(text, default=default)
        try:
            idx = int(response) - 1
            return choices[idx]
        except (ValueError, IndexError):
            print("Invalid choice\n")


def promptyn(msg, default=None):
    """Display a blocking prompt until the user confirms"""
    while True:
        yes = "Y" if default else "y"
        if default or default is None:
            no = "n"
        else:
            no = "N"
        confirm = prompt("%s [%s/%s]" % (msg, yes, no), "").lower()
        if confirm in ("y", "yes"):
            return True
        elif confirm in ("n", "no"):
            return False
        elif not confirm and default is not None:
            return default


def bucket_validate(name):
    """Check for valid bucket name"""
    if name.startswith("."):
        print("Bucket names cannot start with '.'")
        return False
    if name.endswith("."):
        print("Bucket names cannot end with '.'")
        return False
    if ".." in name:
        print("Bucket names cannot contain '..'")
        return False
    return True


def storage_account_name_validate(name):
    """Check for valid storage account name"""
    if "." in name:
        print("Storage account names cannot contain '.'")
        return False
    return True


def make_config(argv=None):
    """Create a server config file"""
    if argv is None:
        argv = sys.argv[1:]
    parser = argparse.ArgumentParser(description=make_config.__doc__)
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "-d", action="store_true", help="Create config file for development"
    )
    group.add_argument("-t", action="store_true", help="Create config file for testing")
    group.add_argument(
        "-p", action="store_true", help="Create config file for production"
    )
    group.add_argument(
        "-r", action="store_true", help="Create config file for docker image"
    )

    parser.add_argument(
        "outfile", nargs="?", help="Name of output file (default stdout)"
    )

    args = parser.parse_args(argv)

    if args.outfile is not None and os.path.exists(args.outfile):
        msg = "'%s' already exists. Overwrite?" % args.outfile
        if not promptyn(msg, False):
            return

    if args.d:
        env = "dev"
    elif args.t:
        env = "test"
    elif args.p:
        env = "prod"
    elif args.r:
        env = "docker"
    else:
        env = prompt_option("What is this config file for?", ["dev", "test", "prod"])

    data = {
        "env": env,
        "workdir": "/var/lib/pypicloud" if env == "docker" else r"%(here)s",
    }
    data["reload_templates"] = env == "dev"

    storage = prompt_option(
        "Where do you want to store your packages?",
        ["s3", "gcs", "filesystem", "azure-blob"],
    )
    if storage == "filesystem":
        storage = "file"

    data["storage"] = storage

    if storage == "s3":
        if "AWS_ACCESS_KEY_ID" in os.environ:
            data["access_key"] = os.environ["AWS_ACCESS_KEY_ID"]
        else:
            data["access_key"] = prompt("AWS access key id?")
        if "AWS_SECRET_ACCESS_KEY" in os.environ:
            data["secret_key"] = os.environ["AWS_SECRET_ACCESS_KEY"]
        else:
            data["secret_key"] = prompt("AWS secret access key?")

        data["s3_bucket"] = prompt("S3 bucket name?", validate=bucket_validate)
        if "." in data["s3_bucket"]:
            data["bucket_region"] = prompt("S3 bucket region?")

    if storage == "gcs":
        data["gcs_bucket"] = prompt("GCS bucket name?", validate=bucket_validate)

    if storage == "azure-blob":
        data["storage_account_name"] = prompt(
            "Storage account name?", validate=storage_account_name_validate
        )
        data["storage_account_key"] = prompt("Storage account key?")
        data["storage_container_name"] = prompt("Container name?")

    data["encrypt_key"] = b64encode(os.urandom(32)).decode("utf-8")
    data["validate_key"] = b64encode(os.urandom(32)).decode("utf-8")

    data["admin"] = prompt("Admin username?")
    data["password"] = _gen_password(_get_password())

    data["session_secure"] = env == "prod"
    data["env"] = env

    if env == "dev" or env == "test":
        data["wsgi"] = "waitress"
    else:
        if hasattr(sys, "real_prefix"):
            data["venv"] = sys.prefix
        data["wsgi"] = "uwsgi"

    tmpl_str = resource_string("pypicloud", "templates/config.ini.jinja2").decode(
        "utf-8"
    )
    template = Template(tmpl_str)

    config_file = template.render(**data)
    if args.outfile is None:
        sys.stdout.write(config_file)
        sys.stdout.write(os.linesep)
    else:
        with open(args.outfile, "w") as ofile:
            ofile.write(config_file)

        print("Config file written to '%s'" % args.outfile)


def migrate_packages(argv=None):
    """
    Migrate packages from one storage backend to another

    Create two config.ini files that are configured to use different storage
    backends. All packages will be migrated from the storage backend in the
    first to the storage backend in the second.

    ex: pypicloud-migrate-packages file_config.ini s3_config.ini

    """
    if argv is None:
        argv = sys.argv[1:]
    parser = argparse.ArgumentParser(
        description=migrate_packages.__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument("config_from", help="Name of config file to migrate from")
    parser.add_argument("config_to", help="Name of config file to migrate to")

    args = parser.parse_args(argv)
    logging.basicConfig()

    old_env = bootstrap(args.config_from)

    old_storage = old_env["request"].db.storage

    all_packages = old_storage.list()

    new_env = bootstrap(args.config_to)
    new_storage = new_env["request"].db.storage
    for package in all_packages:
        print("Migrating %s" % package)
        with old_storage.open(package) as data:
            # we need to recalculate the path for the new storage config
            package.data.pop("path", None)
            new_storage.upload(package, data)


def export_access(argv=None):
    """Dump the access control data to a universal format"""
    if argv is None:
        argv = sys.argv[1:]
    parser = argparse.ArgumentParser(description=export_access.__doc__)
    parser.add_argument("config", help="Name of config file")
    parser.add_argument("-o", help="Name of output file")

    args = parser.parse_args(argv)
    logging.basicConfig()

    env = bootstrap(args.config)
    access = env["request"].access
    data = access.dump()
    if args.o:
        with gzip.open(args.o, "w") as ofile:
            json.dump(data, ofile)
    else:
        print(json.dumps(data, indent=2))


def import_access(argv=None):
    """
    Load the access control data from a dump file or stdin

    This operation is idempotent and graceful. It will not clobber your
    existing ACL.

    """
    if argv is None:
        argv = sys.argv[1:]
    parser = argparse.ArgumentParser(
        description=import_access.__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument("config", help="Name of config file")
    parser.add_argument("-i", help="Name of input file")

    args = parser.parse_args(argv)
    logging.basicConfig()

    if args.i:
        with gzip.open(args.i, "r") as ifile:
            data = json.load(ifile)
    else:
        print("Reading data from stdin...")
        data = json.load(sys.stdin)

    env = bootstrap(args.config)
    access = env["request"].access
    result = access.load(data)
    transaction.commit()
    if result is not None:
        print(result)



""" S3-backed pypi server """
import calendar
import datetime
import logging
from urllib.parse import urlencode

import distlib.locators
from pyramid.config import Configurator
from pyramid.renderers import JSON, render
from pyramid.settings import asbool
from pyramid_beaker import session_factory_from_settings

from .locator import FormattedScrapingLocator, SimpleJsonLocator
from .route import Root
from .util import EnvironSettings

__version__ = "1.3.2"
LOG = logging.getLogger(__name__)


def to_json(value):
    """A json filter for jinja2"""
    return render("json", value)


def _app_url(request, *paths, fragment="", **params):
    """Get the base url for the root of the app plus an optional path"""
    path = "/".join(paths)
    if not path.startswith("/"):
        path = "/" + path
    if params:
        path += "?" + urlencode(params)
    if fragment:
        fragment = "#" + fragment
    return request.application_url + path + fragment


def _fallback_simple(request):
    """/simple endpoint of the fallback url"""
    if request.registry.fallback_url is not None:
        return request.registry.fallback_url
    else:
        return "/".join([request.registry.fallback_base_url, "simple"])


def _add_postfork_hook(config, hook):
    """Add a postfork hook"""
    config.registry.postfork_hooks.append(hook)


def includeme(config):
    """Set up and configure the pypicloud app"""
    config.set_root_factory(Root)
    settings = config.get_settings()
    if getattr(config.registry, "settings", None) is not settings:
        LOG.warning(
            "Settings are not where we expected. Fetching settings from environment variables is disabled. Please file an issue with pypicloud"
        )
    else:
        settings = EnvironSettings(settings)
        config.registry.settings = settings
    config.add_route("health", "/health")
    config.include("pyramid_tm")
    # Beaker should be set by default to invalidate corrupt sessions, otherwise
    # a bad cookie will break the website for you and the only fix is to
    # manually delete the cookie.
    settings.setdefault("session.invalidate_corrupt", "true")
    config.include("pyramid_beaker")
    config.include("pyramid_duh")
    config.include("pyramid_duh.auth")
    config.include("pyramid_rpc.xmlrpc")

    # Sometimes we need to run things after uWSGI forks.
    config.registry.postfork_hooks = []
    config.add_directive("add_postfork_hook", _add_postfork_hook)
    try:
        from uwsgidecorators import postfork
    except ImportError:
        pass
    else:

        @postfork
        def run_postfork_hooks():
            """Run hooks after uWSGI forks"""
            for fn in config.registry.postfork_hooks:
                fn()

    config.include("pypicloud.auth")
    config.include("pypicloud.access")
    config.include("pypicloud.cache")

    # If we're reloading templates, we should also pretty-print json
    reload_templates = asbool(settings.get("pyramid.reload_templates"))
    indent = 4 if reload_templates else None
    json_renderer = JSON(indent=indent)
    json_renderer.add_adapter(
        datetime.datetime, lambda obj, r: calendar.timegm(obj.utctimetuple())
    )
    config.add_renderer("json", json_renderer)
    # Jinja2 configuration
    settings["jinja2.filters"] = {
        "static_url": "pyramid_jinja2.filters:static_url_filter",
        "tojson": to_json,
    }
    settings["jinja2.directories"] = ["pypicloud:templates"]
    config.include("pyramid_jinja2")

    # BEAKER CONFIGURATION
    settings.read_prefix_from_environ("session.")
    config.registry.secure_cookie = asbool(settings.get("session.secure", False))
    settings.setdefault("session.type", "cookie")
    settings.setdefault("session.httponly", "true")
    settings.setdefault("session.crypto_type", "cryptography")
    config.set_session_factory(session_factory_from_settings(settings))

    # PYPICLOUD SETTINGS
    if "pypi.fallback_base_url" in settings:
        config.registry.fallback_base_url = settings["pypi.fallback_base_url"].rstrip(
            "/"
        )
        config.registry.fallback_url = None
    elif "pypi.fallback_url" in settings:
        LOG.warning(
            "pypi.fallback_url is deprecated and support will be "
            "removed in a future version. "
            "Please use pypi.fallback_base_url."
        )
        config.registry.fallback_url = settings["pypi.fallback_url"]
        config.registry.fallback_base_url = None
    else:
        config.registry.fallback_base_url = "https://pypi.org"
        config.registry.fallback_url = None
    config.add_request_method(_fallback_simple, name="fallback_simple", reify=True)

    fallback_mode = settings.get("pypi.fallback", "redirect")
    always_show_upstream = settings.get("pypi.always_show_upstream")

    # Using fallback=mirror is the same as fallback=cache and
    # always_show_upstream=true
    if always_show_upstream is None:
        always_show_upstream = fallback_mode == "mirror"
    else:
        always_show_upstream = asbool(always_show_upstream)

    modes = ("redirect", "cache", "none")
    if fallback_mode not in modes:
        raise ValueError(
            "Invalid value for 'pypi.fallback'. " "Must be one of %s" % ", ".join(modes)
        )
    config.registry.fallback = fallback_mode
    config.registry.always_show_upstream = always_show_upstream

    config.registry.stream_files = asbool(settings.get("pypi.stream_files", False))

    package_max_age = int(settings.get("pypi.package_max_age", 0))
    config.registry.package_max_age = package_max_age

    # Distlib vs JSON scraper
    force_json_scraper = settings.get("pypi.use_json_scraper", None)
    if force_json_scraper is None:
        use_json_scraper = config.registry.fallback_base_url == "https://pypi.org"
    else:
        use_json_scraper = asbool(force_json_scraper)
    if use_json_scraper:
        if config.registry.fallback_base_url is None:
            raise Exception(
                "If setting pypi.use_json_scraper, must also provide pypi.fallback_base_url"
            )
        config.registry.locator = SimpleJsonLocator(config.registry.fallback_base_url)
    else:
        config.registry.locator = FormattedScrapingLocator(_fallback_simple(config))
        LOG.warning(
            "Using distlib scraper. Some packages may fail to install due to "
            "missing 'requires_python' metadata! Set pypi.use_json_scraper = true. "
            "See discussion on https://github.com/stevearc/pypicloud/issues/219"
        )

    # Special request methods
    config.add_request_method(_app_url, name="app_url")
    config.add_request_method(lambda r: r.registry.locator, name="locator", reify=True)
    config.add_request_method(
        lambda x: __version__, name="pypicloud_version", reify=True
    )
    config.add_request_method(
        lambda x: settings.get("pypi.download_url"),
        name="custom_download_url",
        reify=True,
    )

    cache_max_age = int(settings.get("pyramid.cache_max_age", 3600))
    config.add_static_view(
        name="static/%s" % __version__,
        path="pypicloud:static",
        cache_max_age=cache_max_age,
    )

    config.add_xmlrpc_endpoint(
        "pypi", "/pypi", request_method="POST", header="Content-Type:text/xml"
    )
    config.add_xmlrpc_endpoint(
        "pypi_slash", "/pypi/", request_method="POST", header="Content-Type:text/xml"
    )


def main(config, **settings):
    """This function returns a Pyramid WSGI application."""
    config = Configurator(settings=settings)
    config.include("pypicloud")
    config.scan("pypicloud.views")
    return config.make_wsgi_app()



import datetime


class _UTC(datetime.tzinfo):
    def dst(self, dt):
        return datetime.timedelta(0)

    def fromutc(self, dt):
        if dt.tzinfo is None:
            return dt.replace(tzinfo=self)
        return super().fromutc(dt)

    def tzname(self, dt):
        return "UTC"

    def utcoffset(self, dt):
        return datetime.timedelta(0)

    def __repr__(self):
        return "<UTC>"

    def __str__(self):
        return "UTC"


UTC = _UTC()


def utcnow() -> datetime.datetime:
    return datetime.datetime.utcnow().replace(tzinfo=UTC)


def utcfromtimestamp(t: float) -> datetime.datetime:
    return datetime.datetime.utcfromtimestamp(t).replace(tzinfo=UTC)



""" Utilities """
import logging
import os
import re
import time
import unicodedata
from typing import (
    Any,
    Callable,
    Dict,
    ItemsView,
    Iterator,
    KeysView,
    List,
    Optional,
    Tuple,
    Union,
)

from distlib.locators import Locator
from distlib.util import split_filename
from distlib.wheel import Wheel

LOG = logging.getLogger(__name__)
ALL_EXTENSIONS = Locator.source_extensions + Locator.binary_extensions
SENTINEL = object()


class PackageParseError(ValueError):
    pass


def parse_filename(filename: str, name: Optional[str] = None) -> Tuple[str, str]:
    """Parse a name and version out of a filename"""
    version = None
    for ext in ALL_EXTENSIONS:
        if filename.endswith(ext):
            if ext == ".whl":
                wheel = Wheel(filename)
                return wheel.name, wheel.version
            trimmed = filename[: -len(ext)]
            parsed = split_filename(trimmed, name)
            if parsed is None:
                break
            else:
                parsed_name, version = parsed[:2]
            break
    if version is None:
        raise PackageParseError("Cannot parse package file '%s'" % filename)
    if name is None:
        name = parsed_name
    return normalize_name(name), version


def get_packagetype(name: str) -> str:
    """Get package type out of a filename"""
    if name.endswith(".tar.gz"):
        return "sdist"
    elif name.endswith(".egg"):
        return "bdist_egg"
    elif name.endswith(".whl"):
        return "bdist_wheel"
    else:
        return ""


def normalize_name(name: str) -> str:
    """Normalize a python package name"""
    # Lifted directly from PEP503:
    # https://www.python.org/dev/peps/pep-0503/#id4
    return re.sub(r"[-_.]+", "-", name).lower()


def normalize_metadata_value(value: Union[str, bytes]) -> str:
    """Strip non-ASCII characters from metadata values"""
    if isinstance(value, bytes):
        value = value.decode("utf-8")
    if isinstance(value, str):
        value = "".join(c for c in unicodedata.normalize("NFKD", value) if ord(c) < 128)
    return value


def normalize_metadata(metadata: Dict[str, Union[str, bytes]]) -> Dict[str, str]:
    """
    Strip non-ASCII characters from metadata values
    and replace "_" in metadata keys to "-"
    """
    return {
        key.replace("_", "-"): normalize_metadata_value(value)
        for key, value in metadata.items()
    }


def create_matcher(queries: List[str], query_type: str) -> Callable[[str], bool]:
    """
    Create a matcher for a list of queries

    Parameters
    ----------
    queries : list
        List of queries

    query_type: str
        Type of query to run: ["or"|"and"]

    Returns
    -------
        Matcher function

    """
    queries = [query.lower() for query in queries]
    if query_type == "or":
        return lambda x: any((q in x.lower() for q in queries))
    else:
        return lambda x: all((q in x.lower() for q in queries))


class EnvironSettings:
    def __init__(self, settings: Dict[str, Any], env: Dict[str, str] = None):
        self._settings = settings
        if env is None:
            self._env = dict(os.environ)
        else:
            self._env = env

    @staticmethod
    def _get_environ_key(key: str) -> str:
        return "PPC_" + key.upper().replace(".", "_")

    def __iter__(self) -> Iterator[str]:
        return iter(self._settings)

    def keys(self) -> KeysView[str]:
        return self._settings.keys()

    def items(self) -> ItemsView[str, Any]:
        return self._settings.items()

    def __contains__(self, key: str) -> bool:
        env_key = self._get_environ_key(key)
        return env_key in self._env or key in self._settings

    def __setitem__(self, key: str, value: str) -> None:
        env_key = self._get_environ_key(key)
        self._env.pop(env_key, None)
        self._settings[key] = value

    def __getitem__(self, key: str) -> str:
        env_key = self._get_environ_key(key)
        if env_key in self._env:
            return self._env[env_key]
        return self._settings[key]

    def __str__(self) -> str:
        return str(self._settings)

    def __repr__(self) -> str:
        return "EnvironSettings(%s)" % self._settings

    def read_prefix_from_environ(self, prefix: str) -> None:
        if not prefix:
            return
        # Make sure the prefix has a '.' at the end (e.g. "sqlalchemy.")
        if prefix[-1] == "_":
            prefix = prefix[:-1] + "."
        elif prefix[-1] != ".":
            prefix = prefix + "."
        env_prefix = self._get_environ_key(prefix)
        for key, val in self._env.items():
            if key.startswith(env_prefix):
                setting_key = prefix + key[len(env_prefix) :].lower()
                self._settings[setting_key] = val

    def get(self, key: str, default: Any = None) -> Any:
        try:
            return self[key]
        except KeyError:
            return default

    def pop(self, key: str, default: Any = SENTINEL) -> Any:
        env_key = self._get_environ_key(key)
        if env_key in self._env:
            self._settings.pop(key, None)
            return self._env.pop(env_key)
        try:
            return self._settings.pop(key)
        except KeyError:
            if default is SENTINEL:
                raise
            else:
                return default

    def setdefault(self, key: str, value: Any) -> Any:
        try:
            return self[key]
        except KeyError:
            self._settings[key] = value
            return value

    def get_as_dict(
        self, prefix: str, **kwargs: Callable[[Any], Any]
    ) -> Dict[str, Any]:
        """
        Convenience method for fetching settings

        Returns a dict; any settings that were missing from the config file will
        not be present in the returned dict (as opposed to being present with a
        None value)

        Parameters
        ----------
        prefix : str
            String to prefix all keys with when fetching value from settings
        **kwargs : dict
            Mapping of setting name to conversion function (e.g. str or asbool)

        """
        computed = {}
        for name, fxn in kwargs.items():
            val = self.get(prefix + name)
            if val is not None:
                computed[name] = fxn(val)
        return computed

    def clone(self) -> "EnvironSettings":
        return EnvironSettings(dict(self._settings), dict(self._env))


class TimedCache(dict):
    """
    Dict that will store entries for a given time, then evict them

    Parameters
    ----------
    cache_time : int or None
        The amount of time to cache entries for, in seconds. 0 will not cache.
        None will cache forever.
    factory : callable, optional
        If provided, when the TimedCache is accessed and has no value, it will
        attempt to populate itself by calling this function with the key it was
        accessed with. This function should return a value to cache, or None if
        no value is found.

    """

    def __init__(
        self, cache_time: Optional[int], factory: Optional[Callable[[Any], Any]] = None
    ):
        super(TimedCache, self).__init__()
        if cache_time is not None and cache_time < 0:
            raise ValueError("cache_time cannot be negative")
        self._cache_time = cache_time
        self._factory = factory
        self._times = {}  # type: Dict[str, float]

    def _has_expired(self, key):
        """Check if a key is both present and expired"""
        if key not in self._times or self._cache_time is None:
            return False
        updated = self._times[key]
        return updated is not None and time.time() - updated > self._cache_time

    def _evict(self, key):
        """Remove a key if it has expired"""
        if self._has_expired(key):
            del self[key]

    def __contains__(self, key):
        self._evict(key)
        return super(TimedCache, self).__contains__(key)

    def __delitem__(self, key):
        del self._times[key]
        super(TimedCache, self).__delitem__(key)

    def __setitem__(self, key, value):
        if self._cache_time == 0:
            return
        self._times[key] = time.time()
        super(TimedCache, self).__setitem__(key, value)

    def __getitem__(self, key):
        self._evict(key)
        try:
            value = super(TimedCache, self).__getitem__(key)
        except KeyError:
            if self._factory:
                value = self._factory(key)
                if value is None:
                    raise
                self[key] = value
            else:
                raise
        return value

    def get(self, key, default=None):
        self._evict(key)
        value = super(TimedCache, self).get(key, SENTINEL)
        if value is SENTINEL:
            if self._factory is not None:
                value = self._factory(key)
                if value is not None:
                    self[key] = value
                    return value
                else:
                    return default
            else:
                return default
        else:
            return value

    def set_expire(self, key, value, expiration):
        """
        Set a value in the cache with a specific expiration

        Parameters
        ----------
        key : str
        value : value
        expiration : int or None
            Sets the value to expire this many seconds from now. If None, will
            never expire.

        """
        if expiration is not None:
            if expiration <= 0:
                try:
                    del self[key]
                except KeyError:
                    pass
                return
            expiration = time.time() + expiration - self._cache_time

        self._times[key] = expiration
        super(TimedCache, self).__setitem__(key, value)



""" Store package data in redis """
import calendar
import json
import logging
from collections import defaultdict

from pyramid.settings import asbool

from pypicloud.dateutil import utcfromtimestamp, utcnow

from .base import ICache

LOG = logging.getLogger(__name__)


def summary_from_package(package):
    """Create a summary dict from a package"""
    return {
        "name": package.name,
        "summary": package.summary or "",
        "last_modified": package.last_modified,
    }


class RedisCache(ICache):

    """Caching database that uses redis"""

    redis_prefix = "pypicloud:"

    def __init__(self, request=None, db=None, graceful_reload=False, **kwargs):
        super(RedisCache, self).__init__(request, **kwargs)
        self.db = db
        self.graceful_reload = graceful_reload

    @classmethod
    def configure(cls, settings):
        kwargs = super(RedisCache, cls).configure(settings)
        try:
            from redis import StrictRedis
        except ImportError as e:  # pragma: no cover
            raise ImportError(
                "You must 'pip install redis' before using " "redis as the database"
            ) from e
        kwargs["graceful_reload"] = asbool(settings.get("db.graceful_reload", False))
        db_url = settings.get("db.url")
        kwargs["db"] = StrictRedis.from_url(db_url, decode_responses=True)
        return kwargs

    def redis_key(self, key):
        """Get the key to a redis hash that stores a package"""
        return "%spackage:%s" % (self.redis_prefix, key)

    @property
    def redis_set(self):
        """Get the key to the redis set of package names"""
        return self.redis_prefix + "set"

    def redis_filename_set(self, name):
        """Get the key to a redis set of filenames for a package"""
        return "%sset:%s" % (self.redis_prefix, name)

    def redis_summary_key(self, name):
        """Get the redis key to a summary for a package"""
        return "%ssummary:%s" % (self.redis_prefix, name)

    def fetch(self, filename):
        data = self.db.hgetall(self.redis_key(filename))
        if not data:
            return None
        return self._load(data)

    def _load(self, data):
        """Load a Package class from redis data"""
        name = data.pop("name")
        version = data.pop("version")
        filename = data.pop("filename")
        last_modified = utcfromtimestamp(float(data.pop("last_modified")))
        summary = data.pop("summary")
        if summary == "":
            summary = None
        kwargs = dict(((k, json.loads(v)) for k, v in data.items()))
        return self.new_package(
            name, version, filename, last_modified, summary, **kwargs
        )

    def all(self, name):
        filenames = self.db.smembers(self.redis_filename_set(name))
        pipe = self.db.pipeline()
        for filename in filenames:
            pipe.hgetall(self.redis_key(filename))
        packages = [self._load(data) for data in pipe.execute()]
        packages.sort(reverse=True)
        return packages

    def distinct(self):
        return list(self.db.smembers(self.redis_set))

    def summary(self):
        return self._load_summaries(self.db.smembers(self.redis_set))

    def _load_summaries(self, package_names):
        """Load summaries for provided package names"""
        pipe = self.db.pipeline()
        for name in package_names:
            pipe.hgetall(self.redis_summary_key(name))
        summaries = [s for s in pipe.execute() if s]
        for summary in summaries:
            if summary.get("summary", "") == "":
                summary["summary"] = None
            summary["last_modified"] = utcfromtimestamp(float(summary["last_modified"]))
        return summaries

    def clear(self, package):
        count = self._delete_package(package)
        if count == 0:
            self._delete_summary(package.name)

    def _delete_package(self, package, pipe=None):
        """Delete package keys from redis"""
        should_execute = False
        if pipe is None:
            should_execute = True
            pipe = self.db.pipeline()
        pipe.delete(self.redis_key(package.filename))
        pipe.srem(self.redis_filename_set(package.name), package.filename)
        if should_execute:
            pipe.scard(self.redis_filename_set(package.name))
            return pipe.execute()[2]

    def _delete_summary(self, package_name, pipe=None):
        """Delete summary keys from redis"""
        should_execute = False
        if pipe is None:
            should_execute = True
            pipe = self.db.pipeline()
        pipe.srem(self.redis_set, package_name)
        pipe.delete(self.redis_summary_key(package_name))
        if should_execute:
            pipe.execute()

    def clear_all(self):
        keys = self.db.keys(self.redis_prefix + "*")
        if keys:
            self.db.delete(*keys)

    def save(self, package, pipe=None, save_summary=True):
        should_execute = False
        if pipe is None:
            pipe = self.db.pipeline()
            should_execute = True
        dt = package.last_modified
        last_modified = calendar.timegm(dt.utctimetuple()) + dt.microsecond / 1000000.0
        data = {
            "name": package.name,
            "version": package.version,
            "filename": package.filename,
            "last_modified": last_modified,
            "summary": package.summary or "",
        }
        for key, value in package.data.items():
            data[key] = json.dumps(value)
        pipe.hmset(self.redis_key(package.filename), data)
        pipe.sadd(self.redis_set, package.name)
        pipe.sadd(self.redis_filename_set(package.name), package.filename)
        if save_summary:
            self._save_summary(summary_from_package(package), pipe)
        if should_execute:
            pipe.execute()

    def _save_summary(self, summary, pipe):
        """Save a summary dict to redis"""
        dt = summary["last_modified"]
        last_modified = calendar.timegm(dt.utctimetuple()) + dt.microsecond / 1000000.0
        pipe.hmset(
            self.redis_summary_key(summary["name"]),
            {
                "name": summary["name"],
                "summary": summary["summary"] or "",
                "last_modified": last_modified,
            },
        )

    def _load_all_packages(self):
        """Load all packages that are in redis"""
        pipe = self.db.pipeline()
        for filename_key in self.db.keys(self.redis_key("*")):
            pipe.hgetall(filename_key)
        return [self._load(data) for data in pipe.execute() if data]

    def reload_from_storage(self, clear=True):
        if not self.graceful_reload:
            if clear:
                self.clear_all()
            packages = self.storage.list(self.new_package)
            pipe = self.db.pipeline()
            for pkg in packages:
                self.save(pkg, pipe=pipe)
            pipe.execute()
            return

        LOG.info("Rebuilding cache from storage")
        # Log start time
        start = utcnow()
        # Fetch packages from storage s1
        s1 = set(self.storage.list(self.new_package))
        # Fetch cache packages c1
        c1 = set(self._load_all_packages())

        # Add missing packages to cache (s1 - c1)
        missing = s1 - c1
        if missing:
            LOG.info("Adding %d missing packages to cache", len(missing))
            pipe = self.db.pipeline()
            for package in missing:
                self.save(package, pipe, save_summary=False)
            pipe.execute()

        # Delete extra packages from cache (c1 - s1) when last_modified < start
        # The time filter helps us avoid deleting packages that were
        # concurrently uploaded.
        extra1 = [p for p in (c1 - s1) if p.last_modified < start]
        if extra1:
            LOG.info("Removing %d extra packages from cache", len(extra1))
            pipe = self.db.pipeline()
            for package in extra1:
                self._delete_package(package, pipe)
            pipe.execute()

        # If any packages were concurrently deleted during the cache rebuild,
        # we can detect them by polling storage again and looking for any
        # packages that were present in s1 and are missing from s2
        s2 = set(self.storage.list(self.new_package))
        # Delete extra packages from cache (s1 - s2)
        extra2 = s1 - s2
        if extra2:
            LOG.info(
                "Removing %d packages from cache that were concurrently "
                "deleted during rebuild",
                len(extra2),
            )
            pipe = self.db.pipeline()
            for package in extra2:
                self._delete_package(package, pipe)
            pipe.execute()
            # Remove these concurrently-deleted files from the list of packages
            # that were missing from the cache. Don't want to use those to
            # update the summaries below.
            missing -= extra2

        # Update the summary for added packages
        packages_by_name = defaultdict(list)
        for package in missing:
            package.last_modified = package.last_modified
            packages_by_name[package.name].append(package)

        summaries = self._load_summaries(packages_by_name.keys())
        summaries_by_name = {}
        for summary in summaries:
            summaries_by_name[summary["name"]] = summary
        for name, packages in packages_by_name.items():
            if name in summaries_by_name:
                summary = summaries_by_name[name]
            else:
                summary = summary_from_package(packages[0])
                summaries.append(summary)
            for package in packages:
                if package.last_modified > summary["last_modified"]:
                    summary["last_modified"] = package.last_modified
                    summary["summary"] = package.summary
        if summaries:
            LOG.info("Updating %d package summaries", len(summaries))
            pipe = self.db.pipeline()
            for summary in summaries:
                self._save_summary(summary, pipe)
            pipe.execute()

        # Remove the PackageSummary for deleted packages
        removed = set()
        for package in extra1:
            removed.add(package.name)
        for package in extra2:
            removed.add(package.name)
        if removed:
            pipe = self.db.pipeline()
            for name in removed:
                pipe.scard(self.redis_filename_set(name))
            counts = pipe.execute()
            pipe = self.db.pipeline()
            for name, count in zip(removed, counts):
                if count == 0:
                    self._delete_summary(name, pipe)
            pipe.execute()

    def check_health(self):
        from redis import RedisError

        try:
            self.db.echo("ok")
        except RedisError as e:
            return (False, str(e))
        else:
            return (True, "")



""" Store packages as files on disk """
import json
import os
from binascii import hexlify
from contextlib import closing

from pyramid.response import FileResponse

from pypicloud.dateutil import utcfromtimestamp
from pypicloud.models import Package

from .base import IStorage


class FileStorage(IStorage):

    """Stores package files on the filesystem"""

    def __init__(self, request=None, **kwargs):
        self.directory = kwargs.pop("directory")
        super(FileStorage, self).__init__(request, **kwargs)

    @classmethod
    def configure(cls, settings):
        kwargs = super(FileStorage, cls).configure(settings)
        directory = os.path.abspath(settings["storage.dir"]).rstrip("/")
        if not os.path.exists(directory):
            os.makedirs(directory)
        kwargs["directory"] = directory
        return kwargs

    def get_path(self, package):
        """Get the fully-qualified file path for a package"""
        return os.path.join(
            self.directory, package.name, package.version, package.filename
        )

    def path_to_meta_path(self, path):
        """Construct the filename for a metadata file"""
        return path + ".meta"

    def get_metadata_path(self, package):
        """Get the fully-qualified file path for a package metadata file"""
        return self.path_to_meta_path(self.get_path(package))

    def list(self, factory=Package):
        for root, _, files in os.walk(self.directory):
            for filename in files:
                if filename.endswith(".meta"):
                    # We don't want to yield for this file
                    continue

                shortpath = root[len(self.directory) :].strip("/")
                name, version = shortpath.split("/")
                fullpath = os.path.join(root, filename)
                last_modified = utcfromtimestamp(os.path.getmtime(fullpath))
                metadata = {}
                metafile = self.path_to_meta_path(fullpath)
                if os.path.exists(metafile):
                    with open(metafile, "r") as mfile:
                        try:
                            metadata = json.load(mfile)
                        except ValueError:
                            # If JSON fails to decode, don't sweat it.
                            pass
                yield factory(name, version, filename, last_modified, **metadata)

    def download_response(self, package):
        return FileResponse(
            self.get_path(package),
            request=self.request,
            content_type="application/octet-stream",
        )

    def upload(self, package, datastream):
        destfile = self.get_path(package)
        dest_meta_file = self.get_metadata_path(package)
        destdir = os.path.dirname(destfile)
        if not os.path.exists(destdir):
            os.makedirs(destdir)
        uid = hexlify(os.urandom(4)).decode("utf-8")

        # Store metadata as JSON. This could be expanded in the future
        # to store additional metadata about a package (i.e. author)
        tempfile = os.path.join(destdir, ".metadata." + uid)
        with open(tempfile, "w") as mfile:
            json.dump(package.get_metadata(), mfile)

        os.rename(tempfile, dest_meta_file)

        # Write to a temporary file
        tempfile = os.path.join(destdir, "." + package.filename + "." + uid)
        with open(tempfile, "wb") as ofile:
            for chunk in iter(lambda: datastream.read(16 * 1024), b""):
                ofile.write(chunk)

        os.rename(tempfile, destfile)

    def delete(self, package):
        filename = self.get_path(package)
        meta_file = self.get_metadata_path(package)
        os.unlink(filename)
        try:
            os.unlink(meta_file)
        except OSError:
            # Metadata file may not exist
            pass
        version_dir = os.path.dirname(filename)
        try:
            os.rmdir(version_dir)
        except OSError:
            return
        package_dir = os.path.dirname(version_dir)
        try:
            os.rmdir(package_dir)
        except OSError:
            return

    def open(self, package):
        filename = self.get_path(package)
        return closing(open(filename, "rb"))



""" Render views for logging in and out of the web interface """
from pyramid.httpexceptions import HTTPForbidden, HTTPFound
from pyramid.security import NO_PERMISSION_REQUIRED, forget, remember
from pyramid.view import view_config
from pyramid_duh import argify

from pypicloud.route import Root


@view_config(
    context=Root,
    name="login",
    request_method="GET",
    permission=NO_PERMISSION_REQUIRED,
    subpath=(),
    renderer="login.jinja2",
)
def get_login_page(request):
    """Catch login and redirect to login wall"""
    if request.authenticated_userid is not None:
        # User is logged in and fetching /login, so redirect to /
        return HTTPFound(location=request.app_url())
    return {}


@view_config(
    context=HTTPForbidden, permission=NO_PERMISSION_REQUIRED, renderer="login.jinja2"
)
def do_forbidden(request):
    """Intercept 403's and return 401's when necessary"""
    return request.forbid()


@view_config(
    context=Root,
    name="login",
    request_method="POST",
    subpath=(),
    renderer="json",
    permission=NO_PERMISSION_REQUIRED,
)
@argify
def do_login(request, username, password):
    """Check credentials and log in"""
    if request.access.verify_user(username, password):
        request.response.headers.extend(remember(request, username))
        return {"next": request.app_url()}
    else:
        return HTTPForbidden()


def register_new_user(access, username, password):
    """Register a new user & handle duplicate detection"""
    if access.user_data(username) is not None:
        raise ValueError("User '%s' already exists!" % username)
    if username in access.pending_users():
        raise ValueError("User '%s' has already registered!" % username)
    access.register(username, password)
    if access.need_admin():
        access.approve_user(username)
        access.set_user_admin(username, True)
        return True
    return False


def handle_register_request(request, username, password):
    """Process a request to register a new user"""
    if not request.access.allow_register() and not request.access.need_admin():
        return HTTPForbidden()
    username = username.strip()
    try:
        if len(username) > 100 or len(username) < 1:
            raise ValueError("Username must be between 1 and 100 characters")
        if len(password) > 100:
            raise ValueError("Password cannot exceed 100 characters")
        if register_new_user(request.access, username, password):
            request.response.headers.extend(remember(request, username))
    except ValueError as e:
        request.response.status_code = 400
        return {"code": 400, "message": e.args[0]}
    return request.response


@view_config(
    context=Root,
    name="tokenRegister",
    request_method="PUT",
    subpath=(),
    renderer="json",
    permission=NO_PERMISSION_REQUIRED,
)
@argify
def do_token_register(request, token, password):
    """Consume a signed token and create a new user"""
    username = request.access.validate_signup_token(token)
    if username is None:
        raise ValueError("Invalid token")
    if request.access.user_data(username) is not None:
        raise ValueError("User %s already exists" % username)
    request.access.register(username, password)
    request.access.approve_user(username)
    request.response.headers.extend(remember(request, username))
    return {"next": request.app_url()}


@view_config(
    context=Root,
    name="login",
    request_method="PUT",
    subpath=(),
    renderer="json",
    permission=NO_PERMISSION_REQUIRED,
)
@argify
def register(request, username, password):
    """Check credentials and log in"""
    return handle_register_request(request, username, password)


@view_config(context=Root, name="logout", subpath=())
def logout(request):
    """Delete the user session"""
    request.response.headers.extend(forget(request))
    return HTTPFound(location=request.app_url())



""" Backend that reads access control rules from config file """
import logging

from pyramid.settings import aslist

from .base_json import IJsonAccessBackend

LOG = logging.getLogger(__name__)


class ConfigAccessBackend(IJsonAccessBackend):

    """Access Backend that uses values set in the config file"""

    def __init__(self, request=None, data=None, **kwargs):
        super(ConfigAccessBackend, self).__init__(request, **kwargs)
        self._data = data

    @classmethod
    def configure(cls, settings):
        kwargs = super(ConfigAccessBackend, cls).configure(settings)
        data = {}

        users = {}
        for key, value in settings.items():
            if not key.startswith("user."):
                continue
            users[key[len("user.") :]] = value
        data["users"] = users

        data["admins"] = aslist(settings.get("auth.admins", []))

        groups = {}
        for key, value in settings.items():
            if not key.startswith("group."):
                continue
            groups[key[len("group.") :]] = aslist(value)
        data["groups"] = groups

        packages = {}
        for key, value in settings.items():
            pieces = key.split(".")
            if len(pieces) != 4 or pieces[0] != "package":
                continue
            _, package, mode, entity = pieces
            pkg_perms = packages.setdefault(package, {"users": {}, "groups": {}})
            if mode == "user":
                pkg_perms["users"][entity] = cls._perms_from_short(value)
            elif mode == "group":
                pkg_perms["groups"][entity] = cls._perms_from_short(value)
        data["packages"] = packages

        kwargs["data"] = data
        return kwargs

    def _get_db(self):
        return self._data

    @staticmethod
    def _perms_from_short(value):
        """Convert a 'r' or 'rw' specification to a list of permissions"""
        value = value or ""
        if len(value) > 2:
            return aslist(value)
        perms = []
        if "r" in value:
            perms.append("read")
        if "w" in value:
            perms.append("write")
        return perms

    def load(self, data):
        lines = []
        admins = []
        for user in data["users"]:
            lines.append("user.{username} = {password}".format(**user))
            if user.get("admin"):
                admins.append(user["username"])

        if admins:
            lines.append("auth.admins =")
            for admin in admins:
                lines.append("    {0}".format(admin))

        for group, members in data["groups"].items():
            lines.append("group.{0} =".format(group))
            for member in members:
                lines.append("    {0}".format(member))

        def encode_permissions(perms):
            """Encode a permissions list as the r/rw specification"""
            ret = ""
            if "read" in perms:
                ret += "r"
            if "write" in perms:
                ret += "w"
            return ret

        for package, groups in data["packages"]["groups"].items():
            for group, permissions in groups.items():
                lines.append(
                    "package.{0}.group.{1} = {2}".format(
                        package, group, encode_permissions(permissions)
                    )
                )

        for package, users in data["packages"]["users"].items():
            for user, permissions in users.items():
                lines.append(
                    "package.{0}.user.{1} = {2}".format(
                        package, user, encode_permissions(permissions)
                    )
                )

        return "\n".join(lines)



"""LDAP authentication plugin for pypicloud."""
import logging
from collections import namedtuple
from functools import wraps

from pyramid.settings import asbool, aslist

from pypicloud.util import TimedCache

from .base import IAccessBackend
from .config import ConfigAccessBackend

try:
    import ldap
except ImportError as e:  # pragma: no cover
    raise ImportError(
        "You must 'pip install pypicloud[ldap]' before using ldap as the "
        "authentication backend."
    ) from e


LOG = logging.getLogger(__name__)


def reconnect(func):
    """
    If the LDAP connection dies underneath us, recreate it
    """

    @wraps(func)
    def _reconnect(self, *args, **kwargs):
        """
        Inner wrap function to reconnect on failure
        """
        try:
            return func(self, *args, **kwargs)
        except ldap.LDAPError:
            self.connect()
            return func(self, *args, **kwargs)

    return _reconnect


User = namedtuple("User", ["username", "dn", "is_admin"])


class LDAP(object):
    """Handles interactions with the remote LDAP server"""

    def __init__(
        self,
        admin_field,
        admin_group_dn,
        admin_value,
        base_dn,
        cache_time,
        service_dn,
        service_password,
        service_username,
        url,
        user_search_filter,
        user_dn_format,
        ignore_cert,
        ignore_referrals,
        ignore_multiple_results,
    ):
        self._url = url
        self._service_dn = service_dn
        self._service_password = service_password
        self._base_dn = base_dn
        self._user_search_filter = user_search_filter
        self._user_dn_format = user_dn_format
        if user_dn_format is not None:
            if base_dn is not None or user_search_filter is not None:
                raise ValueError(
                    "Cannot use user_dn_format with base_dn " "and user_search_filter"
                )
        else:
            if base_dn is None or user_search_filter is None:
                raise ValueError(
                    "Must provide user_dn_format or both base_dn "
                    "and user_search_filter"
                )
        self._admin_field = admin_field
        self._admin_group_dn = admin_group_dn
        if admin_group_dn and not self._user_dn_format:
            raise ValueError(
                "ldap.admin_group_dn must be used with ldap.user_dn_format"
            )
        self._admin_value = set(admin_value)
        self._server = None
        if cache_time is not None:
            cache_time = int(cache_time)
        self._cache = TimedCache(cache_time, self._fetch_user)
        if service_username is not None:
            self._cache.set_expire(
                service_username, User(service_username, service_dn, True), None
            )
        self._ignore_cert = ignore_cert
        self._ignore_referrals = ignore_referrals
        self._ignore_multiple_results = ignore_multiple_results
        self._admin_member_type = None

    def connect(self):
        """Initializes the python-ldap module and does the initial bind"""
        if self._ignore_cert:
            ldap.set_option(ldap.OPT_X_TLS_REQUIRE_CERT, ldap.OPT_X_TLS_NEVER)
        if self._ignore_referrals:
            ldap.set_option(ldap.OPT_REFERRALS, ldap.OPT_OFF)
        LOG.debug("LDAP connecting to %s", self._url)
        self._server = ldap.initialize(self._url, bytes_mode=False)
        self._bind_to_service()

    @property
    def admin_member_type(self):
        if self._admin_member_type is None:
            LOG.debug("Fetching admin group %s", self._admin_group_dn)
            try:
                results = self._server.search_s(
                    self._admin_group_dn, ldap.SCOPE_BASE, attrlist=["objectClass"]
                )
            except ldap.NO_SUCH_OBJECT as e:
                LOG.debug("NO_SUCH_OBJECT %s", e)
                return "member"
            dn, attributes = results[0]
            classes = [self._decode_attribute(x) for x in attributes["objectClass"]]
            if "groupOfUniqueNames" in classes:
                self._admin_member_type = "uniqueMember"
            else:
                self._admin_member_type = "member"
        return self._admin_member_type

    def _bind_to_service(self):
        """Bind to the service account or anonymous"""
        if self._service_dn:
            # bind with the service_dn
            self._server.simple_bind_s(self._service_dn, self._service_password)
        else:
            # force a connection without binding
            self._server.whoami_s()

    @reconnect
    def test_connection(self):
        """Binds to service. Will throw if bad connection"""
        self._bind_to_service()

    @reconnect
    def _fetch_user(self, username):
        """Fetch a user entry from the LDAP server"""
        LOG.debug("LDAP fetching user %s", username)
        search_attrs = []
        if self._admin_field is not None:
            search_attrs.append(self._admin_field)
        if self._user_dn_format is not None:
            dn = self._user_dn_format.format(username=username)
            LOG.debug("LDAP searching user %r with dn %r", username, dn)
            try:
                results = self._server.search_s(
                    dn, ldap.SCOPE_BASE, attrlist=search_attrs
                )
            except ldap.NO_SUCH_OBJECT as e:
                LOG.debug("NO_SUCH_OBJECT %s", e)
                return
        else:
            search_filter = self._user_search_filter.format(username=username)
            LOG.debug("LDAP searching user %r with filter %r", username, search_filter)
            try:
                results = self._server.search_s(
                    self._base_dn, ldap.SCOPE_SUBTREE, search_filter, search_attrs
                )
            except ldap.NO_SUCH_OBJECT as e:
                LOG.debug("NO_SUCH_OBJECT %s", e)
                return
            except ldap.NO_RESULTS_RETURNED as e:
                LOG.debug("NO_RESULTS_RETURNED %s", e)
                return
        if not results:
            LOG.debug("LDAP user %r not found", username)
            return None
        if len(results) > 1:
            err_msg = "More than one user found for %r: %r" % (
                username,
                [r[0] for r in results],
            )
            if self._ignore_multiple_results:
                LOG.warning(err_msg)
            else:
                raise ValueError(err_msg)
        dn, attributes = results[0]
        LOG.debug("dn: %r, attributes %r", dn, attributes)

        is_admin = False
        if self._admin_field is not None:
            if self._admin_field in attributes:
                is_admin = bool(
                    self._admin_value.intersection(
                        [
                            self._decode_attribute(x)
                            for x in attributes[self._admin_field]
                        ]
                    )
                )
        if not is_admin and self._admin_group_dn:
            user_dn = self._user_dn_format.format(username=username)
            search_filter = "(%s=%s)" % (self.admin_member_type, user_dn)
            LOG.debug(
                "Searching admin group %s for %s", self._admin_group_dn, search_filter
            )
            try:
                results = self._server.search_s(
                    self._admin_group_dn, ldap.SCOPE_BASE, search_filter
                )
            except ldap.NO_SUCH_OBJECT as e:
                LOG.debug("NO_SUCH_OBJECT %s", e)
            else:
                is_admin = bool(results)

        return User(username, dn, is_admin)

    def _decode_attribute(self, attribute):
        if attribute and hasattr(attribute, "decode"):
            decoded = attribute.decode("utf-8")
            return decoded
        else:
            return attribute

    def get_user(self, username):
        """Get the User object or None"""
        return self._cache.get(username)

    @reconnect
    def verify_user(self, username, password):
        """
        Attempts to bind as the user, then rebinds as service user again
        """
        LOG.debug("LDAP verifying user %s", username)
        # Empty password may successfully complete an anonymous bind.
        # Explicitly disallow empty passwords.
        if password == "":
            return False
        user = self._cache.get(username)
        if user is None:
            return False

        try:
            LOG.debug("LDAP binding user %r", user.dn)
            self._server.simple_bind_s(user.dn, password)
        except ldap.INVALID_CREDENTIALS:
            return False
        else:
            return True
        finally:
            self._bind_to_service()


class LDAPAccessBackend(IAccessBackend):
    """
    This backend allows you to authenticate against a remote LDAP server.
    """

    def __init__(self, request=None, conn=None, fallback_factory=None, **kwargs):
        super(LDAPAccessBackend, self).__init__(request, **kwargs)
        self.conn = conn
        self._fallback = None
        self._fallback_factory = fallback_factory

    @property
    def fallback(self):
        if self._fallback is None and self._fallback_factory is not None:
            self._fallback = self._fallback_factory(self.request)
        return self._fallback

    @classmethod
    def configure(cls, settings):
        kwargs = super(LDAPAccessBackend, cls).configure(settings)
        conn = LDAP(
            admin_field=settings.get("auth.ldap.admin_field"),
            admin_group_dn=settings.get("auth.ldap.admin_group_dn"),
            admin_value=aslist(
                settings.get("auth.ldap.admin_value", []), flatten=False
            ),
            base_dn=settings.get("auth.ldap.base_dn"),
            cache_time=settings.get("auth.ldap.cache_time"),
            service_dn=settings.get("auth.ldap.service_dn"),
            service_password=settings.get("auth.ldap.service_password", ""),
            service_username=settings.get("auth.ldap.service_username"),
            url=settings["auth.ldap.url"],
            user_dn_format=settings.get("auth.ldap.user_dn_format"),
            user_search_filter=settings.get("auth.ldap.user_search_filter"),
            ignore_cert=asbool(settings.get("auth.ldap.ignore_cert")),
            ignore_referrals=asbool(settings.get("auth.ldap.ignore_referrals", False)),
            ignore_multiple_results=asbool(
                settings.get("auth.ldap.ignore_multiple_results", False)
            ),
        )
        conn.connect()
        kwargs["conn"] = conn

        fallback = settings.get("auth.ldap.fallback")
        if fallback == "config":
            kw = ConfigAccessBackend.configure(settings)
            kwargs["fallback_factory"] = lambda r: ConfigAccessBackend(r, **kw)

        return kwargs

    def _get_password_hash(self, *_):  # pragma: no cover
        raise RuntimeError("LDAP should never call _get_password_hash")

    def verify_user(self, username, password):
        return self.conn.verify_user(username, password)

    def groups(self, username=None):
        if self.fallback is not None:
            return self.fallback.groups(username)
        # LDAP doesn't support groups by default
        return []

    def group_members(self, group):
        if self.fallback is not None:
            return self.fallback.group_members(group)
        # LDAP doesn't support groups by default
        return []

    def is_admin(self, username):
        if not username:
            return False
        user = self.conn.get_user(username)
        return user is not None and user.is_admin

    def group_permissions(self, package):
        if self.fallback is not None:
            return self.fallback.group_permissions(package)
        return {}

    def user_permissions(self, package):
        if self.fallback is not None:
            return self.fallback.user_permissions(package)
        return {}

    def user_package_permissions(self, username):
        if self.fallback is not None:
            return self.fallback.user_package_permissions(username)
        return []

    def group_package_permissions(self, group):
        if self.fallback is not None:
            return self.fallback.group_package_permissions(group)
        return []

    def user_data(self, username=None):
        if username is None:
            if self.fallback is not None:
                return self.fallback.user_data()
            return []
        else:
            return {
                "username": username,
                "admin": self.is_admin(username),
                "groups": self.groups(username),
            }

    def check_health(self):
        try:
            self.conn.test_connection()
        except ldap.LDAPError as e:
            return (False, str(e))
        else:
            return (True, "")



""" Model objects """
import re
from functools import total_ordering

import pkg_resources

from .dateutil import UTC, utcnow
from .util import normalize_name

METADATA_FIELDS = ["requires_python", "summary", "hash_sha256", "hash_md5"]


@total_ordering
class Package(object):

    """
    Representation of a versioned package

    Parameters
    ----------
    name : str
        The name of the package (will be normalized)
    version : str
        The version number of the package
    filename : str
        The name of the package file
    last_modified : datetime, optional
        The datetime when this package was uploaded (default now)
    summary : str, optional
        The summary of the package
    **kwargs :
        Metadata about the package

    """

    def __init__(
        self, name, version, filename, last_modified=None, summary=None, **kwargs
    ):
        self.name = normalize_name(name)
        self.version = version
        self._parsed_version = None
        self.filename = filename
        if last_modified is not None:
            self.last_modified = last_modified
        else:
            self.last_modified = utcnow()
        if self.last_modified.tzinfo is None:
            self.last_modified = self.last_modified.replace(tzinfo=UTC)
        # Disallow empty string
        self.summary = summary or None
        # Filter out None or empty string
        self.data = {k: v for k, v in kwargs.items() if v}

    def get_url(self, request):
        """Create path to the download link"""
        return request.db.get_url(self)

    @property
    def parsed_version(self):
        """Parse and cache the version using pkg_resources"""
        # Use getattr because __init__ isn't called by some ORMs.
        if getattr(self, "_parsed_version", None) is None:
            self._parsed_version = pkg_resources.parse_version(self.version)
        return self._parsed_version

    @property
    def is_prerelease(self):
        """Returns True if the version is a prerelease version"""
        return re.match(r"^\d+(\.\d+)*$", self.version) is None

    @staticmethod
    def read_metadata(blob):
        """Read metadata from a blob"""
        metadata = {}
        for field in METADATA_FIELDS:
            value = blob.get(field)
            if value:
                metadata[field] = value
                continue
            value = blob.get(field.replace("_", "-"))
            if value:
                metadata[field] = value
        return metadata

    def get_metadata(self):
        """Returns the package metadata as a dict"""
        metadata = Package.read_metadata(self.data)
        if self.summary:
            metadata["summary"] = self.summary
        return metadata

    def __hash__(self):
        return hash(self.filename)

    def __eq__(self, other):
        return isinstance(other, Package) and self.filename == other.filename

    def __lt__(self, other):
        return (self.name, self.parsed_version, self.filename) < (
            other.name,
            other.parsed_version,
            self.filename,
        )

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return "Package(%s)" % self.filename

    def __json__(self, request):
        return {
            "name": self.name,
            "filename": self.filename,
            "last_modified": self.last_modified,
            "version": self.version,
            "url": self.get_url(request),
            "summary": self.summary,
        }

    def search_summary(self):
        """Data to return from a pip search"""
        return {
            "name": self.name,
            "summary": self.summary or "",  # May be None
            "version": self.version,
        }



""" Views for simple pip interaction """
import logging
import posixpath

import pkg_resources
from pyramid.httpexceptions import HTTPBadRequest, HTTPConflict, HTTPFound, HTTPNotFound
from pyramid.view import view_config
from pyramid_duh import addslash, argify
from pyramid_rpc.xmlrpc import xmlrpc_method

from pypicloud.route import Root, SimplePackageResource, SimpleResource
from pypicloud.util import (
    PackageParseError,
    get_packagetype,
    normalize_name,
    parse_filename,
)

LOG = logging.getLogger(__name__)


@view_config(context=Root, request_method="POST", subpath=(), renderer="json")
@view_config(context=SimpleResource, request_method="POST", subpath=(), renderer="json")
@argify
def upload(
    request, content, name=None, version=None, summary=None, requires_python=None
):
    """Handle update commands"""
    action = request.param(":action", "file_upload")
    # Direct uploads from the web UI go here, and don't have a name/version
    if name is None or version is None:
        name, version = parse_filename(content.filename)
    else:
        name = normalize_name(name)
    if action == "file_upload":
        if not request.access.has_permission(name, "write"):
            return request.forbid()
        try:
            return request.db.upload(
                content.filename,
                content.file,
                name=name,
                version=version,
                summary=summary,
                requires_python=requires_python or None,
            )
        except ValueError as e:
            return HTTPConflict(*e.args)
    else:
        return HTTPBadRequest("Unknown action '%s'" % action)


@xmlrpc_method(endpoint="pypi")
@xmlrpc_method(endpoint="pypi_slash")
def search(request, criteria, query_type):
    """
    Perform searches from pip. This handles XML RPC requests to the "pypi"
    endpoint (configured as /pypi/) that specify the method "search".

    """
    filtered = []
    for pkg in request.db.search(criteria, query_type):
        if request.access.has_permission(pkg.name, "read"):
            filtered.append(pkg.search_summary())
    return filtered


@view_config(
    context=SimpleResource, request_method="GET", subpath=(), renderer="simple.jinja2"
)
@addslash
def simple(request):
    """Render the list of all unique package names"""
    names = request.db.distinct()
    i = 0
    while i < len(names):
        name = names[i]
        if not request.access.has_permission(name, "read"):
            del names[i]
            continue
        i += 1
    return {"pkgs": names}


def _package_versions(context, request):
    """Render the links for all versions of a package"""
    fallback = request.registry.fallback
    if fallback == "redirect":
        if request.registry.always_show_upstream:
            return _simple_redirect_always_show(context, request)
        else:
            return _simple_redirect(context, request)
    elif fallback == "cache":
        if request.registry.always_show_upstream:
            return _simple_cache_always_show(context, request)
        else:
            return _simple_cache(context, request)
    else:
        return _simple_serve(context, request)


@view_config(
    context=SimplePackageResource,
    request_method="GET",
    subpath=(),
    renderer="package.jinja2",
)
@addslash
def package_versions(context, request):
    """Render the links for all versions of a package"""
    return _package_versions(context, request)


@view_config(
    context=SimplePackageResource,
    name="json",
    request_method="GET",
    subpath=(),
    renderer="json",
)
def package_versions_json(context, request):
    """Render the package versions in JSON format"""
    pkgs = _package_versions(context, request)
    if not isinstance(pkgs, dict):
        return pkgs
    response = {
        "info": {
            "name": context.name,
            "license": "",
            "classifiers": [],
        },
        "releases": {},
    }
    max_version = None
    for filename, pkg in pkgs["pkgs"].items():
        try:
            name, version_str = parse_filename(filename)
        except PackageParseError:
            continue
        version = pkg_resources.parse_version(version_str)
        if max_version is None or version > max_version:
            max_version = version

        release = {
            "filename": filename,
            "packagetype": get_packagetype(filename),
            "url": pkg.get("non_hashed_url", pkg["url"]),
            "requires_python": pkg["requires_python"],
        }
        if pkg.get("hash_sha256"):
            release["digests"] = {"md5": pkg["hash_md5"], "sha256": pkg["hash_sha256"]}
            release["md5_digest"] = pkg["hash_md5"]

        response["releases"].setdefault(version_str, []).append(release)
    if max_version is not None:
        response["info"]["version"] = str(max_version)
        response["urls"] = response["releases"].get(str(max_version), [])
    return response


def get_fallback_packages(request, package_name, redirect=True):
    """Get all package versions for a package from the fallback_base_url"""
    releases = request.locator.get_releases(package_name)
    pkgs = {}
    if not request.access.has_permission(package_name, "fallback"):
        return pkgs
    for release in releases:
        url = release["url"]
        filename = posixpath.basename(url)
        if not redirect:
            url = request.app_url("api", "package", release["name"], filename)
        pkgs[filename] = {
            "url": url,
            "requires_python": release["requires_python"],
            "hash_sha256": release["digests"].get("sha256"),
            "hash_md5": release["digests"].get("md5"),
        }
    return pkgs


def packages_to_dict(request, packages):
    """Convert a list of packages to a dict used by the template"""
    pkgs = {}
    for package in packages:
        url = package.get_url(request)
        # We could also do with a url without the sha256 fragment for the JSON api
        non_fragment_url = url
        if "#sha256=" in url:
            non_fragment_url = non_fragment_url[: url.find("#sha256=")]

        pkgs[package.filename] = {
            "url": url,
            "non_hashed_url": non_fragment_url,
            "requires_python": package.data.get("requires_python"),
            "hash_sha256": package.data.get("hash_sha256"),
            "hash_md5": package.data.get("hash_md5"),
        }
    return pkgs


def _pkg_response(pkgs):
    """Take a package mapping and return either a dict for jinja or a 404"""
    if pkgs:
        return {"pkgs": pkgs}
    else:
        return HTTPNotFound("No packages found")


def _redirect(context, request):
    """Return a 302 to the fallback url for this package"""
    if request.registry.fallback_base_url:
        path = request.path.lstrip("/")
        redirect_url = "%s/%s" % (request.registry.fallback_base_url.rstrip("/"), path)
    else:
        redirect_url = "%s/%s/" % (
            request.registry.fallback_url.rstrip("/"),
            context.name,
        )

    return HTTPFound(location=redirect_url)


def _simple_redirect(context, request):
    """Service /simple with fallback=redirect"""
    normalized_name = normalize_name(context.name)
    packages = request.db.all(normalized_name)
    if packages:
        if not request.access.has_permission(normalized_name, "read"):
            if request.is_authenticated:
                return _redirect(context, request)
            else:
                return request.request_login()
        else:
            return _pkg_response(packages_to_dict(request, packages))
    else:
        return _redirect(context, request)


def _simple_redirect_always_show(context, request):
    """Service /simple with fallback=redirect"""
    normalized_name = normalize_name(context.name)
    packages = request.db.all(normalized_name)
    if packages:
        if not request.access.has_permission(normalized_name, "read"):
            if request.is_authenticated:
                return _redirect(context, request)
            else:
                return request.request_login()
        else:
            pkgs = get_fallback_packages(request, context.name)
            stored_pkgs = packages_to_dict(request, packages)
            # Overwrite existing package urls
            for filename, url in stored_pkgs.items():
                pkgs[filename] = url
            return _pkg_response(pkgs)
    else:
        return _redirect(context, request)


def _simple_cache(context, request):
    """Service /simple with fallback=cache"""
    normalized_name = normalize_name(context.name)

    if not request.access.has_permission(normalized_name, "read"):
        if request.is_authenticated:
            return HTTPNotFound("No packages found named %r" % normalized_name)
        else:
            return request.request_login()

    packages = request.db.all(normalized_name)
    if packages:
        return _pkg_response(packages_to_dict(request, packages))

    if not request.access.can_update_cache():
        if request.is_authenticated:
            return HTTPNotFound("No packages found named %r" % normalized_name)
        else:
            return request.request_login()
    else:
        pkgs = get_fallback_packages(request, context.name, False)
        return _pkg_response(pkgs)


def _simple_cache_always_show(context, request):
    """Service /simple with fallback=mirror"""
    normalized_name = normalize_name(context.name)

    if not request.access.has_permission(normalized_name, "read"):
        if request.is_authenticated:
            return _redirect(context, request)
        else:
            return request.request_login()

    packages = request.db.all(normalized_name)
    if packages:
        if not request.access.can_update_cache():
            if request.is_authenticated:
                pkgs = get_fallback_packages(request, context.name)
                stored_pkgs = packages_to_dict(request, packages)
                # Overwrite existing package urls
                for filename, data in stored_pkgs.items():
                    pkgs[filename] = data
                return _pkg_response(pkgs)
            else:
                return request.request_login()
        else:
            pkgs = get_fallback_packages(request, context.name, False)
            stored_pkgs = packages_to_dict(request, packages)
            # Overwrite existing package urls
            for filename, data in stored_pkgs.items():
                pkgs[filename] = data
            return _pkg_response(pkgs)
    else:
        if not request.access.can_update_cache():
            if request.is_authenticated:
                return _redirect(context, request)
            else:
                return request.request_login()
        else:
            pkgs = get_fallback_packages(request, context.name, False)
            return _pkg_response(pkgs)


def _simple_serve(context, request):
    """Service /simple with fallback=none"""
    normalized_name = normalize_name(context.name)

    if not request.access.has_permission(normalized_name, "read"):
        if request.is_authenticated:
            return HTTPNotFound("No packages found named %r" % normalized_name)
        else:
            return request.request_login()

    packages = request.db.all(normalized_name)
    return _pkg_response(packages_to_dict(request, packages))



""" Store packages in S3 """
import logging
from binascii import hexlify
from contextlib import contextmanager
from hashlib import md5
from io import BytesIO
from urllib.request import urlopen

from pyramid.httpexceptions import HTTPFound
from pyramid.settings import asbool

from pypicloud.models import Package

from .base import IStorage

LOG = logging.getLogger(__name__)


class ObjectStoreStorage(IStorage):

    """Storage backend base class containing code that is common between
    supported object stores (S3 / GCS)
    """

    test = False

    def __init__(
        self,
        request=None,
        expire_after=None,
        bucket_prefix=None,
        prepend_hash=None,
        redirect_urls=None,
        sse=None,
        object_acl=None,
        storage_class=None,
        region_name=None,
        public_url=False,
        **kwargs
    ):
        super(ObjectStoreStorage, self).__init__(request, **kwargs)
        self.expire_after = expire_after
        self.bucket_prefix = bucket_prefix
        self.prepend_hash = prepend_hash
        self.redirect_urls = redirect_urls
        self.sse = sse
        self.object_acl = object_acl
        self.storage_class = storage_class
        self.region_name = region_name
        self.public_url = public_url

    def _generate_url(self, package: Package) -> str:
        """Subclasses must implement a method for generating signed URLs to
        the package in the object store
        """
        raise NotImplementedError

    @classmethod
    def package_from_object(cls, obj, factory):
        """Subclasses must implement a method for constructing a Package
        instance from the backend's storage object format
        """
        raise NotImplementedError

    @classmethod
    def _subclass_specific_config(cls, settings, common_config):
        """Method to allow subclasses to extract configuration parameters
        specific to them and not covered in the common configuration
        in this class.
        """
        return {}

    @classmethod
    def configure(cls, settings):
        kwargs = super(ObjectStoreStorage, cls).configure(settings)
        kwargs["expire_after"] = int(settings.get("storage.expire_after", 60 * 60 * 24))
        kwargs["bucket_prefix"] = settings.get("storage.prefix", "")
        kwargs["prepend_hash"] = asbool(settings.get("storage.prepend_hash", True))
        kwargs["object_acl"] = settings.get("storage.object_acl", None)
        kwargs["storage_class"] = storage_class = settings.get("storage.storage_class")
        kwargs["redirect_urls"] = asbool(settings.get("storage.redirect_urls", True))
        kwargs["region_name"] = settings.get("storage.region_name")
        kwargs["public_url"] = asbool(settings.get("storage.public_url"))

        kwargs.update(cls._subclass_specific_config(settings, kwargs))
        return kwargs

    def calculate_path(self, package):
        """Calculates the path of a package"""
        path = package.name + "/" + package.filename
        if self.prepend_hash:
            m = md5()
            m.update(package.filename.encode("utf-8"))
            prefix = hexlify(m.digest()).decode("utf-8")[:4]
            path = prefix + "/" + path
        return path

    def get_path(self, package):
        """Get the fully-qualified bucket path for a package"""
        if "path" not in package.data:
            filename = self.calculate_path(package)
            package.data["path"] = self.bucket_prefix + filename
        return package.data["path"]

    def get_url(self, package):
        if self.redirect_urls:
            return super(ObjectStoreStorage, self).get_url(package)
        else:
            return self._generate_url(package)

    def download_response(self, package):
        return HTTPFound(location=self._generate_url(package))

    @contextmanager
    def open(self, package):
        url = self._generate_url(package)
        handle = urlopen(url)
        try:
            yield BytesIO(handle.read())
        finally:
            handle.close()



""" Base class for all cache implementations """
import hashlib
import logging
import posixpath
from io import BytesIO
from typing import Any, BinaryIO, Callable, Dict, List, Optional, Tuple

from pyramid.settings import asbool

from pypicloud.dateutil import utcfromtimestamp
from pypicloud.models import Package
from pypicloud.storage import get_storage_impl
from pypicloud.util import create_matcher, normalize_name, parse_filename

LOG = logging.getLogger(__name__)


class ICache(object):

    """Base class for a caching database that stores package metadata"""

    def __init__(
        self,
        request=None,
        storage=None,
        allow_overwrite=None,
        calculate_hashes=True,
        allow_delete=True,
    ):
        self.request = request
        self.storage = storage(request)
        self.allow_overwrite = allow_overwrite
        self.allow_delete = allow_delete
        self.calculate_hashes = calculate_hashes

    def new_package(self, *args, **kwargs) -> Package:
        return Package(*args, **kwargs)

    def reload_if_needed(self) -> None:
        """
        Reload packages from storage backend if cache is empty

        This will be called when the server first starts

        """
        if not self.distinct():
            LOG.info("Cache is empty. Rebuilding from storage backend...")
            self.reload_from_storage(False)
            LOG.info("Cache repopulated")

    @classmethod
    def configure(cls, settings):
        """Configure the cache method with app settings"""
        return {
            "storage": get_storage_impl(settings),
            "allow_overwrite": asbool(settings.get("pypi.allow_overwrite", False)),
            "allow_delete": asbool(settings.get("pypi.allow_delete", True)),
            "calculate_hashes": asbool(
                settings.get("pypi.calculate_package_hashes", True)
            ),
        }

    @classmethod
    def postfork(cls, **kwargs):
        """This method will be called after uWSGI forks"""

    def get_url(self, package: Package) -> str:
        """
        Get the download url for a package

        Parameters
        ----------
        package : :class:`~pypicloud.models.Package`

        Returns
        -------
        url : str

        """
        return self.storage.get_url(package)

    def download_response(self, package: Package):
        """Pass through to storage"""
        return self.storage.download_response(package)

    def reload_from_storage(self, clear: bool = True) -> None:
        """Make sure local database is populated with packages"""
        if clear:
            self.clear_all()
        packages = self.storage.list(self.new_package)
        for pkg in packages:
            self.save(pkg)

    def upload(
        self,
        filename: str,
        data: BinaryIO,
        name: Optional[str] = None,
        version: Optional[str] = None,
        summary: Optional[str] = None,
        requires_python: Optional[str] = None,
    ) -> Package:
        """
        Save this package to the storage mechanism and to the cache

        Parameters
        ----------
        filename : str
            Name of the package file
        data : file
            File-like readable object
        name : str, optional
            The name of the package (if not provided, will be parsed from
            filename)
        version : str, optional
            The version number of the package (if not provided, will be parsed
            from filename)
        summary : str, optional
            The summary of the package
        requires_python : str, optional
            The Python version requirement

        Returns
        -------
        package : :class:`~pypicloud.models.Package`
            The Package object that was uploaded

        Raises
        ------
        e : ValueError
            If the package already exists and allow_overwrite = False

        """
        if version is None or name is None:
            name, version = parse_filename(filename, name)
        name = normalize_name(name)
        filename = posixpath.basename(filename)
        old_pkg = self.fetch(filename)
        metadata = {"requires_python": requires_python}
        if old_pkg is not None and not self.allow_overwrite:
            raise ValueError("Package '%s' already exists!" % filename)
        if self.calculate_hashes:
            file_data = data.read()
            metadata["hash_sha256"] = hashlib.sha256(file_data).hexdigest()
            metadata["hash_md5"] = hashlib.md5(file_data).hexdigest()
            data = BytesIO(file_data)

        new_pkg = self.new_package(name, version, filename, summary=summary, **metadata)
        self.storage.upload(new_pkg, data)
        self.save(new_pkg)
        return new_pkg

    def delete(self, package: Package) -> None:
        """
        Delete this package from the database and from storage

        Parameters
        ----------
        package : :class:`~pypicloud.models.Package`

        """
        if not self.allow_delete:
            raise ValueError(
                "Cannot delete packages. Set pypi.allow_delete = true if you want to enable deletes."
            )
        self.storage.delete(package)
        self.clear(package)

    def fetch(self, filename: str) -> Package:
        """
        Get matching package if it exists

        Parameters
        ----------
        filename : str
            Name of the package file

        Returns
        -------
        package : :class:`~pypicloud.models.Package`

        """
        raise NotImplementedError

    def all(self, name: str) -> List[Package]:
        """
        Search for all versions of a package

        Parameters
        ----------
        name : str
            The name of the package

        Returns
        -------
        packages : list
            List of all :class:`~pypicloud.models.Package` s with the given
            name

        """
        raise NotImplementedError

    def distinct(self) -> List[str]:
        """
        Get all distinct package names

        Returns
        -------
        names : list
            List of package names

        """
        raise NotImplementedError

    def search(self, criteria: Dict[str, List[str]], query_type: str) -> List[Package]:
        """
        Perform a search from pip

        Parameters
        ----------
        criteria : dict
            Dictionary containing the search criteria. Pip sends search criteria
            for "name" and "summary" (typically, both of these lists have the
            same search values).

            Example::

                {
                    "name": ["value1", "value2", ..., "valueN"],
                    "summary": ["value1", "value2", ..., "valueN"]
                }

        query_type : str
            Type of query to perform. By default, pip sends "or".


        """
        name_queries = criteria.get("name", [])
        summary_queries = criteria.get("summary", [])
        packages = []

        # Create matchers for the queries
        match_name = create_matcher(name_queries, query_type)
        match_summary = create_matcher(summary_queries, query_type)

        for key in self.distinct():
            # Search all versions of this package key
            latest: Optional[Package] = None
            for package in self.all(key):
                # Search for a match. If we've already found a match, make sure
                # we find the most recent version that matches.
                if latest is None or package > latest:
                    if match_name(package.name):
                        latest = package
                    elif package.summary is not None and match_summary(package.summary):
                        latest = package
            if latest is not None:
                packages.append(latest)

        return packages

    def summary(self) -> List[Dict[str, Any]]:
        """
        Summarize package metadata

        Returns
        -------
        packages : list
            List of package dicts, each of which contains 'name', 'summary',
            and 'last_modified'.

        """
        packages = []
        for name in self.distinct():
            pkg = {
                "name": name,
                "summary": "",
                "last_modified": utcfromtimestamp(0),
            }
            max_pkg: Optional[Package] = None
            for package in self.all(name):
                pkg["last_modified"] = max(pkg["last_modified"], package.last_modified)
                if max_pkg is None:
                    max_pkg = package
                elif package > max_pkg:
                    max_pkg = package
            if max_pkg:
                pkg["summary"] = max_pkg.summary
                packages.append(pkg)

        return packages

    def clear(self, package: Package) -> None:
        """
        Remove this package from the caching database

        Parameters
        ----------
        package : :class:`~pypicloud.models.Package`

        """
        raise NotImplementedError

    def clear_all(self) -> None:
        """Clear all cached packages from the database"""
        raise NotImplementedError

    def save(self, package: Package) -> None:
        """
        Save this package to the database

        Parameters
        ----------
        package : :class:`~pypicloud.models.Package`

        """
        raise NotImplementedError

    def check_health(self) -> Tuple[bool, str]:
        """
        Check the health of the cache backend

        Returns
        -------
        (healthy, status) : (bool, str)
            Tuple that describes the health status and provides an optional
            status message

        """
        return True, ""



""" Backend that defers to another server for access control """
import json
from json import JSONDecodeError

import boto3
from botocore.exceptions import ClientError

from pypicloud.util import EnvironSettings

from .base_json import IMutableJsonAccessBackend


class AWSSecretsManagerAccessBackend(IMutableJsonAccessBackend):

    """
    This backend allows you to store all user and package permissions in
    AWS Secret Manager

    """

    def __init__(
        self, request=None, secret_id=None, kms_key_id=None, client=None, **kwargs
    ):
        super(AWSSecretsManagerAccessBackend, self).__init__(request, **kwargs)
        self.secret_id = secret_id
        self.kms_key_id = kms_key_id
        self.client = client
        self.dirty = False

    @classmethod
    def configure(cls, settings: EnvironSettings):
        kwargs = super(AWSSecretsManagerAccessBackend, cls).configure(settings)
        kwargs["secret_id"] = settings["auth.secret_id"]
        kwargs["kms_key_id"] = settings.get("auth.kms_key_id")
        session = boto3.session.Session(
            **settings.get_as_dict(
                "auth.",
                region_name=str,
                aws_access_key_id=str,
                aws_secret_access_key=str,
                aws_session_token=str,
                profile_name=str,
            )
        )
        kwargs["client"] = session.client("secretsmanager")

        return kwargs

    def _get_db(self):
        """Hit a server endpoint and return the json response"""
        try:
            response = self.client.get_secret_value(SecretId=self.secret_id)
        except ClientError as e:
            if e.response["Error"]["Code"] == "ResourceNotFoundException":
                return {}
            elif e.response["Error"]["Code"] == "InvalidRequestException":
                raise Exception("The request was invalid due to:", e) from e
            elif e.response["Error"]["Code"] == "InvalidParameterException":
                raise Exception("The request had invalid params:", e) from e
            raise

        try:
            return json.loads(response["SecretString"])
        except JSONDecodeError as e:
            raise Exception("Invalid json detected: {}".format(e)) from e

    def _save(self):
        if not self.dirty:
            self.dirty = True
            self.request.tm.get().addAfterCommitHook(self._do_save)

    def _do_save(self, succeeded):
        """Save the auth data to the backend"""
        if not succeeded:
            return
        kwargs = {"SecretString": json.dumps(self._db)}
        if self.kms_key_id is not None:
            kwargs["KmsKeyId"] = self.kms_key_id
        try:
            self.client.update_secret(SecretId=self.secret_id, **kwargs)
        except ClientError as e:
            if e.response["Error"]["Code"] == "ResourceNotFoundException":
                self.client.create_secret(Name=self.secret_id, **kwargs)
            raise

    def check_health(self):
        try:
            self._get_db()
        except Exception as e:
            return (False, str(e))
        else:
            return (True, "")
